HTTP_STATUS(100, CONTINUE, Continue)
HTTP_STATUS(101, SWITCHING_PROTOCOLS, Switching Protocols)
HTTP_STATUS(102, PROCESSING, Processing)
HTTP_STATUS(103, EARLY_HINTS, Early Hints)
HTTP_STATUS(200, OK, OK)
HTTP_STATUS(201, CREATED, Created)
HTTP_STATUS(202, ACCEPTED, Accepted)
HTTP_STATUS(203, NON_AUTHORITATIVE_INFORMATION, Non Authoritative Information)
HTTP_STATUS(204, NO_CONTENT, No Content)
HTTP_STATUS(205, RESET_CONTENT, Reset Content)
HTTP_STATUS(206, PARTIAL_CONTENT, Partial Content)
HTTP_STATUS(207, MULTI_STATUS, Multi Status)
HTTP_STATUS(208, ALREADY_REPORTED, Already Reported)
HTTP_STATUS(218, THIS_IS_FINE, This Is Fine)
HTTP_STATUS(226, IM_USED, IM Used)
HTTP_STATUS(300, MULTIPLE_CHOICES, Multiple Choices)
HTTP_STATUS(301, MOVED_PERMANENTLY, Moved Permanently)
HTTP_STATUS(302, FOUND, Found)
HTTP_STATUS(303, SEE_OTHER, See Other)
HTTP_STATUS(304, NOT_MODIFIED, Not Modified)
HTTP_STATUS(305, USE_PROXY, Use Proxy)
HTTP_STATUS(306, SWITCH_PROXY, Switch Proxy)
HTTP_STATUS(307, TEMPORARY_REDIRECT, Temporary Redirect)
HTTP_STATUS(308, PERMANENT_REDIRECT, Permanent Redirect)
HTTP_STATUS(400, BAD_REQUEST, Bad Request)
HTTP_STATUS(401, UNAUTHORIZED, Unauthorized)
HTTP_STATUS(402, PAYMENT_REQUIRED, Payment Required)
HTTP_STATUS(403, FORBIDDEN, Forbidden)
HTTP_STATUS(404, NOT_FOUND, Not Found)
HTTP_STATUS(405, METHOD_NOT_ALLOWED, Method Not Allowed)
HTTP_STATUS(406, NOT_ACCEPTABLE, Not Acceptable)
HTTP_STATUS(407, PROXY_AUTHENTICATION_REQUIRED, Proxy Authentication Required)
HTTP_STATUS(408, REQUEST_TIMEOUT, Request Timeout)
HTTP_STATUS(409, CONFLICT, Conflict)
HTTP_STATUS(410, GONE, Gone)
HTTP_STATUS(421, MISDIRECTED_REQUEST, Misdirected Request)
HTTP_STATUS(422, UNPROCESSABLE_CONTENT, Unprocessable Content)
HTTP_STATUS(499, CLIENT_CLOSED_REQUEST, Client Closed Request)
HTTP_STATUS(500, INTERNAL_SERVER_ERROR, Internal Server Error)
HTTP_STATUS(501, NOT_IMPLEMENTED, Not Implemented)
HTTP_STATUS(502, BAD_GATEWAY, Bad Gateway)
HTTP_STATUS(503, SERVICE_UNAVAILABLE, Service Unavailable)
HTTP_STATUS(504, GATEWAY_TIMEOUT, Gateway Timeout)
HTTP_STATUS(505, HTTP_VERSION_NOT_SUPPORTED, HTTP Version Not Supported)
HTTP_STATUS(507, INSUFFICIENT_STORAGE, Insufficient Storage)
